/*
 * Copyright 2002 Niels Provos <provos@citi.umich.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Niels Provos.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>

#include "config.h"

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/tree.h>
#include <sys/queue.h>
#include <sys/wait.h>

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <dnet.h>
#include <ctype.h>

#undef timeout_pending
#undef timeout_initialized

#include <event.h>

#include "honeyd.h"
#include "personality.h"

SPLAY_HEAD(perstree, personality) personalities;

int npersons;

struct personate person_drop = {};

int
perscompare(struct personality *a, struct personality *b)
{
	return (strcmp(a->name, b->name));
}

SPLAY_PROTOTYPE(perstree, personality, node, perscompare);

SPLAY_GENERATE(perstree, personality, node, perscompare);

void
personality_init(void)
{
	npersons = 0;
	SPLAY_INIT(&personalities);
}

struct personality *
personality_new(char *name)
{
	struct personality *pers, tmp;

	tmp.name = name;
	if (SPLAY_FIND(perstree, &personalities, &tmp))
		return (NULL);

	if ((pers = calloc(1, sizeof(struct personality))) == NULL)
		err(1, "%s: calloc", __FUNCTION__);

	if ((pers->name = strdup(name)) == NULL)
		err(1, "%s: stdup", __FUNCTION__);

	npersons++;
	SPLAY_INSERT(perstree, &personalities, pers);

	return (pers);
}

struct personality *
personality_clone(struct personality *person)
{
	struct personality *newperson;

	if ((newperson = malloc(sizeof(struct personality))) == NULL)
		err(1, "%s: malloc", __FUNCTION__);

	memcpy(newperson, person, sizeof(struct personality));
	return (newperson);
}

void
personality_free(struct personality *pers)
{
	SPLAY_REMOVE(perstree, &personalities, pers);
	free(pers->name);
	free(pers);
}

struct personality *
personality_random(void)
{
	extern rand_t *honeyd_rand;
	struct personality *pers;
	int i;

	if (!npersons)
		return (NULL);

	i = rand_uint32(honeyd_rand) % npersons;
	pers = SPLAY_MIN(perstree, &personalities);
	while (i--) {
		pers = SPLAY_NEXT(perstree, &personalities, pers);
	}

	return (pers);
}

struct personality *
personality_find(char *name)
{
	struct personality tmp;
	tmp.name = name;
	return (SPLAY_FIND(perstree, &personalities, &tmp));
}

/* Not much here, set up ip id accordingly */

void
ip_personality(struct template *tmpl, uint16_t *pid)
{
	struct personality *person;

	if (tmpl == NULL)
		return;
	if ((person = tmpl->person) == NULL)
		return;

	if (person->idt == ID_SEQUENTIAL)
		*pid = tmpl->id++;

}

struct personate *
tcp_personality_test(struct tcp_con *con, struct personality *person,
    uint8_t sndflags)
{
	uint8_t flags;

	flags = con->rcv_flags & (TH_FIN|TH_RST|TH_PUSH|TH_ACK|TH_URG|TH_SYN);
	if (flags == TH_SYN) {
		switch (con->state) {
		case TCP_STATE_LISTEN:
		case TCP_STATE_SYN_RECEIVED:
			if (sndflags & TH_RST)
				return (NULL);
			return (&person->tests[0]);
		case TCP_STATE_CLOSED:
			return (&person->tests[4]);
		default:
			return (NULL);
		}
	} else if (flags == 0) {
		switch (con->state) {
		case TCP_STATE_LISTEN:
			return (&person->tests[1]);
		default:
			return (NULL);
		}
	} else if (flags == (TH_SYN|TH_PUSH|TH_FIN|TH_URG)) {
		switch (con->state) {
		case TCP_STATE_LISTEN:
		case TCP_STATE_SYN_RECEIVED:
			return (&person->tests[2]);
		default:
			return (NULL);
		}
	} else if (flags == TH_ACK) {
		switch (con->state) {
		case TCP_STATE_LISTEN:
		case TCP_STATE_SYN_RECEIVED:
			return (&person->tests[3]);
		case TCP_STATE_CLOSED:
			return (&person->tests[5]);
		default:
			return (NULL);
		}
	} else if (flags == (TH_FIN|TH_PUSH|TH_URG)) {
		switch (con->state) {
		case TCP_STATE_CLOSED:
			return (&person->tests[6]);
		default:
			return (NULL);
		}
	} else if ((flags & TH_FIN) && (flags & (TH_SYN|TH_ACK)) == 0) {
		if (person->disallow_finscan || con->state != TCP_STATE_LISTEN)
			return (NULL);
		return (&person_drop);
	}

	return (NULL);
}

#define TIME_CORRECT(x) do { \
	tmp.tv_sec = 0; \
	tmp.tv_usec = tmp.tv_usec % (x); \
	timersub(&tmpl->tv, &tmp, &tmpl->tv); \
} while (0)

uint32_t
tcp_personality_seq(struct template *tmpl, struct personality *person)
{
	struct timeval tv, tmp;
	extern rand_t *honeyd_rand;
	int slowhz;
	uint32_t total, num;

	tmpl->seqcalls++;

	if (!timerisset(&tmpl->tv)) {
		gettimeofday(&tmpl->tv, NULL);
		if (tmpl->timestamp == 0)
			tmpl->timestamp = rand_uint32(honeyd_rand) % 1728000;
		if (tmpl->seq == 0)
			tmpl->seq = rand_uint32(honeyd_rand);
		return (tmpl->seq);
	}

	gettimeofday(&tv, NULL);
	timersub(&tv, &tmpl->tv, &tmp);
	tmpl->tv = tv;
	slowhz = (tmp.tv_sec * 2 + tmp.tv_usec / 500000);
	TIME_CORRECT(500000);

	tmpl->timestamp += slowhz;

	switch(person->seqt) {
	case SEQ_CONSTANT:
		return (tmpl->seq);
	case SEQ_TRIVIALTIME:
		tmpl->seq += 32121;

		num = person->seqindex_max - person->seqindex_min;
		if (num)
			num = rand_uint32(honeyd_rand) % num;
		num += person->seqindex_min;
		if (tmpl->seqcalls % 2)
			tmpl->seq -= num;
		else
			tmpl->seq += num;

		return (tmpl->seq);
	case SEQ_RPI: {
		total = slowhz * 1000;
		tmpl->seq += 64000;
		num = person->seqindex_max - person->seqindex_min;
		if (num)
			num = rand_uint32(honeyd_rand) % num;
		num += person->seqindex_min;
		if (tmpl->seqcalls % 2)
			tmpl->seq -= num;
		else
			tmpl->seq += num;
		if (total) {
			tmpl->seq += total/2;
			tmpl->seq += (rand_uint32(honeyd_rand) % (total/4));
		}
		return (tmpl->seq);
	}
	case SEQ_CLASS64K:
		tmpl->seq += (slowhz + 1) * 64000;
		return (tmpl->seq);
	case SEQ_RANDOM:
	default:
		return (rand_uint32(honeyd_rand));
	}
}

char *
tcp_personality(struct tcp_con *con, uint8_t *pflags, int *pwindow, int *pdf,
    uint16_t *pid)
{
	struct template *tmpl = con->tmpl;
	struct personality *person;
	struct personate *pers;

	if (tmpl == NULL)
		return (NULL);
	person = tmpl->person;
	if (person == NULL)
		return (NULL);

	if ((pers = tcp_personality_test(con, person, *pflags)) == NULL)
		return (NULL);

	*pwindow = pers->window;
	*pflags = pers->flags;
	*pdf = pers->df;

	switch (pers->forceack) {
	case ACK_ZERO:
		con->rcv_next = 0;
		break;
	case ACK_DECREMENT:
		con->rcv_next--;
		break;
	case ACK_KEEP:
		break;
	}

	ip_personality(tmpl, pid);

	if (con->snd_una == 0)
		con->snd_una = tcp_personality_seq(tmpl, person);

	return (pers->options);
}

#define SET(y, x, w, l) do { \
	(x)->opt_type = w; \
	(x)->opt_len = l; \
	memcpy(y, x, l); \
} while (0)

void
tcp_personality_options(struct tcp_con *con, struct tcp_hdr *tcp,
    char *options)
{
	extern rand_t *honeyd_rand;
	u_char *p = (u_char *)tcp + TCP_HDR_LEN;
	struct template *tmpl = con->tmpl;
	struct tcp_opt opt;
	int optlen = 0;
	uint32_t timestamp;
	short mss = 1000;
	char *o;
	
	for (o = options; *o; o++) {
		opt.opt_len = 0;
		switch(*o) {
		case 'm':
			if (o[1] == 'e') {
				o++;
				if (con->mss)
					mss = con->mss;
			}
			opt.opt_data.mss = htons(mss);
			SET(p, &opt, TCP_OPT_MSS, 4);
			break;
		case 'w':
			opt.opt_data.wscale = 0;
			SET(p, &opt, TCP_OPT_WSCALE, 3);
			break;
		case 't':
			if (tmpl != NULL)
				timestamp = htonl(tmpl->timestamp);
			else
				timestamp = rand_uint32(honeyd_rand);
			opt.opt_data.timestamp[0] = timestamp;
			opt.opt_data.timestamp[1] = 0;
			if (con->sawtimestamp)
				opt.opt_data.timestamp[1] = con->echotimestamp;
			SET(p, &opt, TCP_OPT_TIMESTAMP, 2 + 4 + 4);
			break;
		case 'n':
			if (!optlen)
				break;
			SET(p, &opt, TCP_OPT_NOP, 1);
			break;
		case 'l':
			SET(p, &opt, TCP_OPT_EOL, 2);
			break;
		default:
			opt.opt_len = 0;
			break;
		}
		optlen += opt.opt_len;
		p += opt.opt_len;
	}

	if (optlen)
		tcp->th_off += (optlen + 3) / 4;
}

#define RVAL_DO(x, w) do { \
	switch (w) { \
	case RVAL_OKAY: break; \
	case RVAL_ZERO: (x) = 0; \
	default: (x) = ntohs(x); \
	} \
} while (0)

int
icmp_personality(struct addr *dst, struct ip_hdr *ip, uint8_t *pdf,
    uint8_t *ptos, int *pquotelen)
{
	struct persudp *test;
	struct template *tmpl;

	tmpl = template_find(addr_ntoa(dst));
	if (tmpl == NULL)
		tmpl = template_find("default");
	if (tmpl == NULL || tmpl->person == NULL)
		return (1);

	test = &tmpl->person->udptest;

	if (!test->response)
		return (0);

	*pdf = test->df;
	*ptos = test->tos;
	*pquotelen = test->quotelen;
	if (test->riplen) {
		u_int len = ntohs(ip->ip_len);
		ip->ip_len = htons(len + test->riplen);
	}
	RVAL_DO(ip->ip_id, test->rid);
	RVAL_DO(ip->ip_sum, test->ripck);
	if (ip->ip_p == IP_PROTO_UDP) {
		struct udp_hdr *udp = (struct udp_hdr *)((u_char *)ip + (ip->ip_hl << 2));
		u_char *p = (u_char *)(udp + 1);
		RVAL_DO(udp->uh_sum, test->uck);
		if (test->dat == RVAL_BAD)
			*p = 0;
	}
	
	return (1);
}

/* Parse nmap tests */

int
parse_tseq(struct personality *pers, int off, char *line)
{
	char *p = line, *p2, *end;


	while (p != NULL && strlen(p)) {
		p2 = strsep(&p, "%");
		/* We ignore all other values, only take the first */
		end = p2;
		p2 = strsep(&end, "|");

		if (strncasecmp(p2, "Class=", 6) == 0) {
			enum seqtype st = 0;
			p2 = line + 6;
	   
			if (strcasecmp(p2, "TD") == 0)
				st = SEQ_TRIVIALTIME;
			else if (strcasecmp(p2, "TR") == 0)
				st = SEQ_RANDOM;
			else if (strcasecmp(p2, "RI") == 0)
				st = SEQ_RPI;
			else if (strcasecmp(p2, "C") == 0)
				st = SEQ_CONSTANT;
			else if (strcasecmp(p2, "64K") == 0)
				st = SEQ_CLASS64K;
			else if (strcasecmp(p2, "i800") == 0)
				st = SEQ_I800;
			else
				return (-1);

			pers->seqt = st;
		} else if (strncasecmp(p2, "SI=", 3) == 0) {
			char *next;
			p2 += 3;

			while (p2 && *p2) {
				next = p2;
				p2 = strsep(&next, "&");

				if (*p2 == '<') {
					p2++;
					pers->seqindex_max = strtoul(p2, &end, 16);
				} else if (*p2 == '>') {
					p2++;
					pers->seqindex_min = strtoul(p2, &end, 16);
				} else {
					uint32_t seqindex;

					seqindex = strtoul(p2, &end, 16);
					pers->seqindex_min = seqindex;
					pers->seqindex_max = seqindex;
				}
				if (end == NULL || *end != '\0')
					return (-1);
				p2 = next;
			}
			if (pers->seqindex_max == 0)
				pers->seqindex_max = 2 * pers->seqindex_min;
			if (pers->seqindex_min > pers->seqindex_max)
				return (-1);
		}
	}

	if (pers->seqt == SEQ_RPI &&
	    !pers->seqindex_min && !pers->seqindex_max)
		pers->seqindex_max = 0x10000;

	return (0);
}

int
parse_tl(struct personality *pers, int off, char *line)
{
	struct personate *test = &pers->tests[off];
	char *p = line, *p2, *end;

	if (strncasecmp(line, "Resp=N", 6) == 0) {
		test->flags = 0;
		return (0);
	}


	/* Permits Y|N, too */
	if (strncasecmp(line, "Resp=Y", 6) == 0) {
		p = strchr(p, '%');
		if (p == NULL)
			return (-1);
		p++;
	}
		
	while (p != NULL && strlen(p)) {
		p2 = strsep(&p, "%");
		/* We ignore all other values, only take the first */
		end = p2;
		p2 = strsep(&end, "|");

		if (strcasecmp(p2, "DF=Y") == 0) {
			test->df = 1;
		} else if (strcasecmp(p2, "DF=N") == 0) {
			test->df = 0;
		} else if (strncasecmp(p2, "W=", 2) == 0) {
			int smaller = 0;
			p2 += 2;

			/* Special cases */
			if (strcasecmp(p2, "O") == 0)
				continue;
			if (*p2 == '<') {
				p2++;
				smaller = 1;
			}

			test->window = strtoul(p2, &end, 16);
			if (end == NULL || *end != '\0')
				return (-1);

			if (smaller)
				test->window--;
		} else if (strncasecmp(p2, "ACK=", 4) == 0) {
			p2 += 4;
			if (strcasecmp(p2, "O") == 0)
				test->forceack = ACK_ZERO;
			else if (strcasecmp(p2, "S") == 0)
				test->forceack = ACK_DECREMENT;
			else if (strcasecmp(p2, "S++") == 0)
				test->forceack = ACK_KEEP;
			else
				return (-1);
		} else if (strncasecmp(p2, "Flags=", 6) == 0) {
			p2 += 6;
			
			test->flags = 0;
			for (; *p2; p2++) {
				*p2 = tolower(*p2);
				switch (*p2) {
				case 'a':
					test->flags |= TH_ACK;
					break;
				case 's':
					test->flags |= TH_SYN;
					break;
				case 'f':
					test->flags |= TH_FIN;
					break;
				case 'r':
					test->flags |= TH_RST;
					break;
				case 'p':
					test->flags |= TH_PUSH;
					break;
				case 'u':
					test->flags |= TH_URG;
					break;
				case 'b':
					test->flags |= TH_CWR;
					break;
				default:
					return (-1);
				}
			}
		} else if (strncasecmp(p2, "Ops=", 4) == 0) {
			char *p3;
			p2 += 4;
			if (strlen(p2)) {
				for (p3 = p2; *p3; p3++)
					*p3 = tolower(*p3);
				if ((test->options = strdup(p2)) == NULL)
					err(1, "%s: strdup", __FUNCTION__);
			}
		} else
		      return (-1);
	}
	
	return (0);
}

#define RVAL_TRANS(x, p) do { \
	if (strcasecmp(p, "0") == 0) \
		(x) = RVAL_ZERO; \
	else if (strcasecmp(p, "E") == 0) \
		(x) = RVAL_OKAY; \
	else if (strcasecmp(p, "F") == 0) \
		(x) = RVAL_BAD; \
	else if (*p == '\0') \
		(x) = RVAL_OKAY;	/* Fill in default */ \
	else \
		return (-1); \
} while (0)

int
parse_pu(struct personality *pers, int off, char *line)
{
	struct persudp *test = &pers->udptest;
	char *p = line, *p2, *end;

	if (strncasecmp(line, "Resp=N", 6) == 0) {
		test->response = 0;
		return (0);
	}
	test->response = 1;

	if (strncasecmp(line, "Resp=Y%", 7) == 0)
		p += 7;
		
	while (p != NULL && strlen(p)) {
		p2 = strsep(&p, "%");
		/* We ignore all other values, only take the first */
		end = p2;
		p2 = strsep(&end, "|");

		if (strcasecmp(p2, "DF=Y") == 0) {
			test->df = 1;
		} else if (strcasecmp(p2, "DF=N") == 0) {
			test->df = 0;
		} else if (strcasecmp(p2, "DF=") == 0) {
			/* Fill in default */
			test->df = 0;
		} else if (strncasecmp(p2, "ULEN=", 5) == 0) {
			continue;
		} else if (strncasecmp(p2, "TOS=", 4) == 0) {
			p2 += 4;

			test->tos = strtoul(p2, &end, 16);
			if (end == NULL || *end != '\0')
				return (-1);
		} else if (strncasecmp(p2, "IPLEN=", 6) == 0) {
			p2 += 6;

			test->quotelen = strtoul(p2, &end, 16);
			if (end == NULL || *end != '\0')
				return (-1);
			test->quotelen -= IP_HDR_LEN + ICMP_HDR_LEN + 4;
		} else if (strncasecmp(p2, "RIPTL=", 6) == 0) {
			p2 += 6;

			test->riplen = strtoul(p2, &end, 16) - 328;
			if (end == NULL || *end != '\0')
				return (-1);
		} else if (strncasecmp(p2, "RID=", 4) == 0) {
			p2 += 4;
			RVAL_TRANS(test->rid, p2);
		} else if (strncasecmp(p2, "RIPCK=", 6) == 0) {
			p2 += 6;
			RVAL_TRANS(test->ripck, p2);
		} else if (strncasecmp(p2, "UCK=", 4) == 0) {
			p2 += 4;
			RVAL_TRANS(test->uck, p2);
		} else if (strncasecmp(p2, "DAT=", 4) == 0) {
			p2 += 4;
			RVAL_TRANS(test->dat, p2);
		} else
			return (-1);
	}
	
	return (0);
}

struct parse_test {
	char *start;
	int offset;
	int (*parse_test)(struct personality *, int, char *);
} parse_tests[] = {
	{"TSeq", 0, parse_tseq},
	{"T1", 0, parse_tl},
	{"T2", 1, parse_tl},
	{"T3", 2, parse_tl},
	{"T4", 3, parse_tl},
	{"T5", 4, parse_tl},
	{"T6", 5, parse_tl},
	{"T7", 6, parse_tl},
	{"PU", 0, parse_pu},
	{NULL, 0, NULL}
};

int
personality_line(struct personality *pers, char *line)
{
	struct parse_test *pt = parse_tests;
	char *p, *p2;

	p2 = line;
	p = strsep(&p2, "(");
	if (p2 == NULL)
		return (-1);
	p = strsep(&p2, ")");
	if (p2 == NULL)
		return (-1);
	for (; pt->start; pt++) {
		if (strncasecmp(line, pt->start, strlen(pt->start)) == 0)
			return (pt->parse_test(pers, pt->offset, p));
	}

	return (-1);
}

/* Parsing Functions */

#define WHITESPACE	" \t\n"
#define FINGERPRINT	"Fingerprint"
#define CMP(y,x)	strncasecmp(y, x, sizeof(x) - 1)

int
personality_parse(FILE *fin)
{
	char bl[1024], line[1024], *p, *p2;
	int errors = 0, lineno = 0, ignore = 0;
	struct personality *pers = NULL;

	while ((p = fgets(line, sizeof(line), fin)) != NULL) {
		strlcpy(bl, line, sizeof(bl));

		lineno++;
		p += strspn(p, WHITESPACE);

		if (*p == '\0' || *p == '#') {
			pers = NULL;
			continue;
		}

		/* Remove trailing comments */
		p2 = p;
		strsep(&p2, "#\r\n");

		if (CMP(p, FINGERPRINT) == 0) {
			p += sizeof(FINGERPRINT) - 1;
			p += strspn(p, ": \t");
			if (!isalnum(*p)) {
				fprintf(stderr, "%d: bad name \"%s\"\n",
				    lineno, p);
				return (-1);
			}
			for (p2 = p + strlen(p) - 1; isblank(*p2); p2--)
				*p2 = '\0';

			if ((pers = personality_new(p)) == NULL) {
				fprintf(stderr,
				    "%d: Ignoring duplicate fingerprint"
				    " \"%s\"\n", lineno, p);
				ignore = 1;
			} else
				ignore = 0;
			continue;
		}

		if (pers == NULL) {
			if (ignore)
				continue;

			fprintf(stderr, "%d: No personality for \"%s\"\n",
			    lineno, p);
			return (-1);
		}

		if (personality_line(pers, p) == -1) {
			fprintf(stderr, "%d: parse error: %s", lineno, bl);
			errors++;
		}
	}

	return (errors ? -1 : 0);
}

